print("import: 'persistent'")
import persistent

print("import: 'persistent._timestamp'")
import persistent._timestamp

print("import: 'persistent.cPersistence'")
import persistent.cPersistence

print("import: 'persistent.cPickleCache'")
import persistent.cPickleCache

